#
/*
 *
 *    Copyright (C) 2010, 2011, 2012
 *    Jan van Katwijk (J.vanKatwijk@gmail.com)
 *    JFF Consultancy
 *
 *    This file is part of the JSDR.
 *    Many of the ideas as implemented in JSDR are derived from
 *    other work, made available through the GNU general Public License. 
 *    All copyrights of the original authors are recognized.
 *
 *    JSDR is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    JSDR is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with JSDR; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include	"fm_processor.h"
#include	"gui.h"
#include	"fm_demodulator.h"
#include	"sincos.h"
#include	"fm-constants.h"
#include	"rds_decoder.h"
#include	"pa_writer.h"

#define	AUDIO_FREQ_DEV_PROPORTION 0.85f
#define	PILOT_FREQUENCY		19000
#define	RDS_FREQUENCY		(3 * PILOT_FREQUENCY)
#define	OMEGA_DEMOD		(1.0 / fmRate) * (2 * M_PI)
#define	OMEGA_PILOT	((DSPFLOAT (PILOT_FREQUENCY)) / fmRate) * (2 * M_PI)
#define	OMEGA_RDS	((DSPFLOAT) RDS_FREQUENCY / fmRate) * (2 * M_PI)

static inline
DSPCOMPLEX makeComplex (uint8_t re, uint8_t im) {
	return DSPCOMPLEX (DSPFLOAT (re - 128) / 128.0,
	                   DSPFLOAT (im - 128) / 128.0);
}

//
//	The fm processor  takes a sample and returns 3 values
//	a. one to show in a spectrum to the world (selectable)
//	b. one containing the audio in two channels encoded in a complex
//	c. one containg the raw data for interpretation by the rds processor
//	Note that no decimation done as yet.
	fmProcessor::fmProcessor (virtualInput		*vi,
	                          RadioInterface	*RI,
	                          paWriter		*mySink,
	                          int32_t		dabRate,
	                          int32_t		fmRate,
	                          int32_t		workingRate,
	                          int32_t		audioRate,
	                          int32_t		displaySize,
	                          int32_t		spectrumSize,
	                          int32_t		averageCount,
	                          int32_t		repeatRate,
	                          RingBuffer<double> *hfBuffer,
	                          RingBuffer<DSPCOMPLEX> *lfBuffer,
	                          SinCos		*mySinCos) {
	running				= false;
	this	-> myRig		= vi;
	this	-> myRadioInterface	= RI;
	this	-> theSink		= mySink;
	this	-> dabRate		= dabRate;
	this	-> fmRate		= fmRate;
	this	-> decimatingScale	= dabRate / fmRate;
	this	-> workingRate		= workingRate;
	this	-> audioRate		= audioRate;
	this	-> displaySize		= displaySize;
	this	-> averageCount		= averageCount;
	this	-> freezer		= 0;
	this	-> repeatRate		= repeatRate;
	this	-> hfBuffer		= hfBuffer;
	this	-> lfBuffer		= lfBuffer;
	this	-> mySinCos		= mySinCos;
	if ((spectrumSize & (spectrumSize - 1)) != 0)
	   this	-> spectrumSize	= 4 * displaySize;
	else
	   this	-> spectrumSize		= spectrumSize;
	this	-> spectrum_fft		= new common_fft (this -> spectrumSize);
	this	-> spectrumBuffer	= spectrum_fft -> getVector ();
	this	-> displayBuffer	= new double [displaySize];
	memset (displayBuffer, 0, displaySize * sizeof (double));

	this	-> localOscillator	= new Oscillator (dabRate);
	this	-> lo_frequency		= 0;
	this	-> omega_demod		= 2 * M_PI / fmRate;
	this	-> fmBandwidth		= 0.95 * fmRate;
	this	-> fmFilterDegree	= 21;
	this	-> fmFilter		= new LowPassFIR (21,
	                                                  0.95 * fmRate / 2,
	                                                  fmRate);
	this	-> newFilter		= false;
/*
 *	default values, will be set through the user interface
 *	to their appropriate values
 */
	this	-> fmModus		= FM_STEREO;
	this	-> selector		= S_STEREO;
	this	-> balance		= 0;
	this	-> leftChannel		= - (balance - 50.0) / 100.0;
	this	-> rightChannel		= (balance + 50.0) / 100.0;
	this	-> Volume		= 0.1;
	this	-> viewSelector		= 1;
	this	-> upFilter		= new LowPassFIR (11,
	                                                  workingRate,
	                                                  audioRate);
	this	-> audioDecimator	= new downDecimator (fmRate,
	                                                     workingRate);
	this	-> rdsDecimator		= new downDecimator (fmRate,
	                                                     workingRate / 2);
	this	-> myRdsDecoder		= new rdsDecoder (myRadioInterface,
	                                                  workingRate / 2,
	                                                  mySinCos);
/*
 *	averagePeakLevel and audioGain are set
 *	prior to calling the processFM method
 */
	this	-> peakLevel		= -100;
	this	-> peakLevelcnt		= 0;
	this	-> max_freq_deviation	= 0.95 * (0.5 * fmRate);
	this	-> norm_freq_deviation	= 0.8 * max_freq_deviation;
	this	-> audioGain		= 1.0;
//
	this	-> fm_Levels		= new fmLevels (LEVEL_SIZE,
	                                                fmRate, LEVEL_FREQ);
//
//	Since data is coming with a pretty high rate, we need to filter
//	and decimate, for which we have a decimating filter
	fmBandfilter		= new DecimatingFIR (decimatingScale * 5,
	                                             fmRate / 2,
	                                             dabRate,
	                                             decimatingScale);
//
//	for isolating the pilot signal, we need a reasonable
//	filter. The filtered signal is beautified by a pll
	pilotBandFilter		= new fftFilter (FFT_SIZE, PILOTFILTER_SIZE);
	pilotBandFilter		-> setBand (PILOT_FREQUENCY - PILOT_WIDTH / 2,
		                            PILOT_FREQUENCY + PILOT_WIDTH / 2,
		                            fmRate);
	pilotRecover		= new pilotRecovery (fmRate,
	                                             OMEGA_PILOT,
	                                             25 * omega_demod,
	                                             mySinCos);
	pilotDelay	= (FFT_SIZE / 2 - PILOTFILTER_SIZE) * OMEGA_PILOT;

	rdsLowPassFilter	= new fftFilter (FFT_SIZE, RDSLOWPASS_SIZE);
	rdsLowPassFilter	-> setLowPass (RDS_WIDTH, fmRate);
//
//	the constant K_FM is still subject to many questions
	DSPFLOAT	F_G	= 65000;	// highest freq in message
	DSPFLOAT	Delta_F	= 75000;	//
	DSPFLOAT	B_FM	= 2 * (Delta_F + F_G);
	K_FM			= 2 * M_PI / 1.0 * (B_FM / 2 - F_G);
	K_FM			= K_FM * 1.0 / fmRate;

	TheDemodulator		= new fm_Demodulator (fmRate,
	                                              mySinCos, K_FM);
	fmAudioFilter		= NULL;
//
//	In the case of mono we do not assume a pilot
//	to be available. We borrow the approach from CuteSDR
	rdsHilbertFilter	= new HilbertFilter (HILBERT_SIZE,
	                                     (DSPFLOAT)RDS_FREQUENCY / fmRate,
	                                             fmRate);
	rdsBandFilter		= new fftFilter (FFT_SIZE,
	                                         RDSBANDFILTER_SIZE);
	rdsBandFilter		-> setSimple (RDS_FREQUENCY - RDS_WIDTH / 2,
	                                      RDS_FREQUENCY + RDS_WIDTH / 2,
	                                      fmRate);
	rds_plldecoder		= new pllC (fmRate,
	                                    RDS_FREQUENCY,
	                                    RDS_FREQUENCY - 50,
	                                    RDS_FREQUENCY + 50,
	                                    200,
	                                    mySinCos);

//	for the deemphasis we use an in-line filter with
	xkm1			= 0;
	ykm1			= 0;
	alpha			= 1.0 / (fmRate / (1000000.0 / 50.0 + 1));
	viewSelector		= 1;
	dumping			= false;
	dumpFile		= NULL;

	connect (this, SIGNAL (hfBufferLoaded (int)),
	         myRadioInterface, SLOT (hfBufferLoaded (int)));
	connect (this, SIGNAL (lfBufferLoaded (int)),
	         myRadioInterface, SLOT (lfBufferLoaded (int)));
	running			= true;
	start ();
}

	fmProcessor::~fmProcessor (void) {
	running	= false;
	while (!isFinished ())
	   usleep (100);
	delete	TheDemodulator;
	delete	rds_plldecoder;
	delete	pilotRecover;
	delete	rdsHilbertFilter;
	delete	rdsBandFilter;
	delete	pilotBandFilter;
	delete	fm_Levels;
	delete	displayBuffer;
	delete	spectrum_fft;
	if (fmAudioFilter != NULL)
	   delete fmAudioFilter;
}

DSPFLOAT	fmProcessor::getPilotStrength	(void) {
	return fm_Levels	-> getPilotStrength ();
}

DSPFLOAT	fmProcessor::getRdsStrength	(void) {
	return fm_Levels	-> getRdsStrength ();
}

DSPFLOAT	fmProcessor::getNoiseStrength	(void) {
	return fm_Levels	-> getNoiseStrength ();
}

DSPFLOAT	fmProcessor::get_DcComponent	(void) {
	return TheDemodulator	-> get_DcComponent ();
}

const char  *fmProcessor::nameofDecoder	(void) {
	return TheDemodulator -> nameofDecoder ();
}

//
//	changing a filter is in two steps: here we set a marker,
//	but the actual filter is created in the mainloop of
//	the processor
void	fmProcessor::setBandwidth	(int32_t b) {
	fmBandwidth	= b;
	newFilter	= true;
}

void	fmProcessor::setBandfilterDegree	(int32_t d) {
	fmFilterDegree	= d;
	newFilter	= true;
}

void	fmProcessor::setfmMode (uint8_t m) {
	fmModus	= m ? FM_STEREO : FM_MONO;
}

void	fmProcessor::setFMdecoder (int8_t d) {
	TheDemodulator	-> setDecoder (d);
}

void	fmProcessor::setView (int16_t d) {
	viewSelector	= d;
}

void	fmProcessor::setSoundMode (uint8_t selector) {
	this	-> selector = selector;
}

void	fmProcessor::setSoundBalance (int16_t balance) {
	this	-> balance = balance;
	leftChannel	= - (balance - 50.0) / 100.0;
	rightChannel	= (balance + 50.0) / 100.0;
}
//	Deemphasis	= 50 usec (3183 Hz, Europe)
//	Deemphasis	= 75 usec (2122 Hz US)
//	tau		= 2 * M_PI * Freq = 1000000 / time
void	fmProcessor::setDeemphasis	(int16_t v) {
DSPFLOAT	Tau;

	switch (v) {
	   default:
	      v	= 1;
	   case	1:
	   case 50:
	   case 75:
//	pass the Tau
	      Tau	= 1000000.0 / v;
	      alpha	= 1.0 / (DSPFLOAT (fmRate) / Tau + 1.0);
	}
}

void	fmProcessor::setVolume (int16_t Vol) {
	Volume = 0.1 * Vol;
}

DSPCOMPLEX	fmProcessor::audioGainCorrection (DSPCOMPLEX z) {
	return cmul (z, audioGain * Volume);
}

void	fmProcessor::startDumping	(FILE *f) {
	if (dumping)
	   return;
//	do not change the order here, another thread might get confused
	dumpFile 	= f;
	dumping		= true;
}

void	fmProcessor::stopDumping	(void) {
	dumping = false;
}

void	fmProcessor::setAttenuation (int16_t l, int16_t r) {
	Lgain	= l;
	Rgain	= r;
}

//
/*
 *	In this variant, we have a real task for processing
 */

void	fmProcessor::run (void) {
DSPCOMPLEX	result;
DSPFLOAT 	rdsData;
int32_t		bufferSize	= 2 * 8192;
uint8_t		*dataBuffer	=
	         (uint8_t *)alloca (bufferSize * 2 * sizeof (uint8_t));
double	*Y_values	= (double *)alloca (displaySize * sizeof (double));
int16_t		i, j;
DSPCOMPLEX	out;
DSPCOMPLEX	pcmSample;
int16_t		hfCount	= 0, lfCount = 0;
//
int16_t		amount = bufferSize < spectrumSize / 2 ? bufferSize : spectrumSize / 2;
	while (running) {
	   int32_t x = myRig -> Samples ();
	   while (x < bufferSize) {
	      usleep (10);
	      x = myRig -> Samples ();
	      if (!running)
	         break;
	   }

	   if (!running)
	      break;

	   if (newFilter && (fmBandwidth < 0.95 * fmRate)) {
	      newFilter = false;
	      delete fmFilter;
	      fmFilter	= new LowPassFIR (fmFilterDegree,
	                                  fmBandwidth / 2,
	                                  fmRate);
	   }

	   (void)myRig -> getSamples (dataBuffer, bufferSize);

//	for the HFscope
	   if (++hfCount > (dabRate / bufferSize) / repeatRate) {
	      for (i = 0; i < amount / 2; i ++)
	         spectrumBuffer [i] = makeComplex (dataBuffer [2 * i],
	                                           dataBuffer [2 * i + 1]);
	      for (i = amount / 2; i < spectrumSize; i ++)
	         spectrumBuffer [i] = 0;

	      spectrum_fft	-> do_FFT ();
//	and map the spectrumSize values onto displaySize elements
	      for (i = 0; i < displaySize / 2; i ++) {
	         double f	= 0;
	         for (j = 0; j < spectrumSize / displaySize; j ++)
	            f += abs (spectrumBuffer [spectrumSize /
	                                             displaySize * i + j]);
	         Y_values [displaySize / 2 + i] = 
                              f / (spectrumSize / displaySize);
	         f = 0;
	         for (j = 0; j < spectrumSize / displaySize; j ++)
	            f += abs (spectrumBuffer [spectrumSize / 2 +
	                             spectrumSize / displaySize * i + j]);
	         Y_values [i] =
	                      f / (spectrumSize / displaySize);
	      }

	      if (freezer > 0) {
	         for (i = 0; i < displaySize; i ++)
	            displayBuffer [i] =
	                1.0 / freezer * Y_values [i] +
	                (freezer - 1.0) / freezer * displayBuffer [i];
	         freezer ++;
	      }
	      else
	         for (i = 0; i < displaySize; i ++)
	            displayBuffer [i] = 
	                1.0 /averageCount * Y_values [i] +
	                (averageCount - 1.0) / averageCount * displayBuffer [i];

	      memcpy (Y_values, displayBuffer, displaySize * sizeof (double));
	      hfBuffer	-> putDataIntoBuffer (Y_values, displaySize);
	      hfCount = 0;
//
//	and signal the GUI thread that we have data
	      emit hfBufferLoaded (displaySize);
	   }

	   if (dumping) 
	      (void)fwrite ((void *)dataBuffer, sizeof (uint8_t),
	                                        2 * bufferSize, dumpFile);
	   for (i = 0; i < bufferSize; i ++) {
	      DSPCOMPLEX v = 
	          DSPCOMPLEX ((dataBuffer [2 * i] - 128) / 128.0 * Lgain,
	                      (dataBuffer [2 * i + 1] - 128) / 128.0 * Rgain);
	      v	= v * localOscillator -> nextValue (lo_frequency);
//
//	first step: decimating (and filtering)
	      if (!fmBandfilter -> Pass (v, &v))
	         continue;
//
//	second step: if requested, apply filtering
	      if (fmBandwidth < 0.95 * fmRate)
	         v	= fmFilter	-> Pass (v);
//
//	Now we have the signal ready for decoding
//	keep track of the peaklevel, we take segments
//	of 1024 samples
	      if (abs (v) > peakLevel)
	         peakLevel = abs (v);
	      if (++peakLevelcnt >= 1024) {
	         DSPFLOAT	ratio	=
	                          max_freq_deviation / norm_freq_deviation;
	         if (peakLevel > 0)
	         this -> audioGain	=
	                  (ratio / peakLevel) / AUDIO_FREQ_DEV_PROPORTION;
	         if (audioGain == 0)
	            audioGain = 1.0;
	         peakLevelcnt	= 0;
	         peakLevel	= -100;
	      }

	      if (fmModus == FM_MONO) {
	         mono (&v, &result, &rdsData);
	      }
	      else {
	         stereo (&v, &result, &rdsData);
	         switch (selector) {
	            default:
	            case S_STEREO:
	               result = DSPCOMPLEX (real (result) + imag (result),
	                                  - (- real (result) + imag (result)));
	               break;

	            case S_LEFT:
	               result = DSPCOMPLEX (real (result) + imag (result), 
	                                    real (result) + imag (result));
	               break;

	            case S_RIGHT:
	               result = DSPCOMPLEX (- (imag (result) - real (result)),
	                                    - (imag (result) - real (result)));
	               break;

	            case S_LEFTplusRIGHT:
	               result = DSPCOMPLEX (real (result),  real (result));
	               break;

	            case S_LEFTminusRIGHT:
	               result = DSPCOMPLEX (imag (result), imag (result));
	               break;
	         }
	      }
//
//	"result" now contains the audio sample, either stereo
//	or mono
	      result = audioGainCorrection (result);
	      if (fmAudioFilter != NULL)
	         result = fmAudioFilter -> Pass (result);
	      out  = DSPCOMPLEX (leftChannel * real (result),
	                         rightChannel * imag (result));

	      if (audioDecimator -> doDecimate (out, &pcmSample))
	         sendSampletoOutput (pcmSample);

	      if ((rdsModus != rdsDecoder::NO_RDS) &&
	           rdsDecimator -> doDecimate (rdsData, &rdsData)) {
	         DSPFLOAT mag;
	         myRdsDecoder -> doDecode (rdsData, &mag,
	                                   (rdsDecoder::RdsMode)rdsModus);
	      }
//	samples for the LF scope are handled bufferwise
	      lfBuffer	-> putDataIntoBuffer (&v, 1);
	      if (++lfCount > 2048) {
	         emit lfBufferLoaded (lfCount);
	         lfCount = 0;
	      }
	   }
	}
}

void	fmProcessor::mono (DSPCOMPLEX	*in,
	                   DSPCOMPLEX	*audioOut,
	                   DSPFLOAT	*rdsValue) {
DSPFLOAT	Re, Im;
DSPCOMPLEX	rdsBase;
DSPFLOAT	demod	= TheDemodulator	-> demodulate (*in);

	fm_Levels	-> addItem (demod);
	if (viewSelector == 1)
	   *in	= demod;
//
//	deemphasize
	Re	= xkm1 = (demod - xkm1) * alpha + xkm1;
	Im	= ykm1 = (demod - ykm1) * alpha + ykm1;
	*audioOut	= DSPCOMPLEX (Re, Im);

	if (viewSelector == 2 || viewSelector == 3)
	   *in = *audioOut;
//
//	fully inspired by cuteSDR, we try to decode the rds stream
//	by simply am decoding it (after creating a decent complex
//	signal by Hilbert filtering)
	rdsBase	= DSPCOMPLEX (5 * demod, 5 * demod);
	rdsBase = rdsHilbertFilter -> Pass (rdsBandFilter -> Pass (rdsBase));
	if (viewSelector == 4)
	   *in = rdsBase;
	rds_plldecoder -> do_pll (rdsBase);
	if (viewSelector == 5)
	   *in = mySinCos -> getComplex (rds_plldecoder -> getNco ());
	DSPFLOAT rdsDelay = imag (rds_plldecoder -> getDelay ());
	if (viewSelector == 6)
	   *in = rdsDelay;
	*rdsValue = rdsLowPassFilter -> Pass (5 * rdsDelay);
	if (viewSelector == 7)
	   *in =  *rdsValue;
}

void	fmProcessor::stereo (DSPCOMPLEX	*in,
	                     DSPCOMPLEX	*audioOut,
	                     DSPFLOAT	*rdsValue) {

DSPFLOAT	LRPlus	= 0;
DSPFLOAT	LRDiff	= 0;
DSPFLOAT	pilot	= 0;
DSPFLOAT	currentPilotPhase;
DSPFLOAT	PhaseforLRDiff	= 0;
DSPFLOAT	PhaseforRds	= 0;
/*
 */
	DSPFLOAT demod	= TheDemodulator  -> demodulate (*in);
	fm_Levels	-> addItem (demod);
	if (viewSelector == 1)
	   *in = demod;
	LRPlus = LRDiff = pilot	= demod;
/*
 *	get the phase for the "carrier to be inserted" right
 */
	pilot		= pilotBandFilter -> Pass (5 * pilot);
	if (viewSelector == 2)
	   *in = pilot;
	currentPilotPhase = pilotRecover -> getPilotPhase (5 * pilot);
	if (viewSelector == 3)
	   *in = mySinCos	-> getComplex (currentPilotPhase);
/*
 *	Now we have the right - i.e. synchronized - signal to work with
 */
	PhaseforLRDiff	= 2 * (currentPilotPhase + pilotDelay);
	PhaseforRds	= 3 * (currentPilotPhase + pilotDelay);

	LRDiff	= 2 * mySinCos	-> getCos (PhaseforLRDiff) * LRDiff;
	if (viewSelector == 4)
	   *in = LRDiff;
//
	DSPCOMPLEX  MixerValue = mySinCos -> getComplex (PhaseforRds);
	if (viewSelector == 5)
	   *in = MixerValue;

	*rdsValue = 5 * rdsLowPassFilter -> Pass (real (MixerValue) * demod);
	if (viewSelector == 6)
	   *in =  *rdsValue;

//	apply deemphasis
	LRPlus		= xkm1	= (LRPlus - xkm1) * alpha + xkm1;
	LRDiff		= ykm1	= (LRDiff - ykm1) * alpha + ykm1;
        *audioOut		= DSPCOMPLEX (LRPlus, LRDiff);
	if (viewSelector == 7)
	   *in = *audioOut;
}

void	fmProcessor::setLFcutoff (int32_t Hz) {
	if (fmAudioFilter != NULL)
	   delete	fmAudioFilter;
	fmAudioFilter	= NULL;
	if (Hz > 0)
	   fmAudioFilter	= new LowPassIIR (10,
	                                          Hz,
	                                          fmRate,
	                                          S_BUTTERWORTH);
}

bool	fmProcessor::isLocked (void) {
	return ((fmModus == FM_STEREO) && (pilotRecover -> isLocked ()));
}

void	fmProcessor::sendSampletoOutput (DSPCOMPLEX s) {
int16_t i;

	if (audioRate == workingRate)
	   theSink	-> putSample (s);
	else {
	   theSink	-> putSample (upFilter -> Pass (s));
	   for (i = 1; i < audioRate / workingRate; i ++)
	      theSink	-> putSample (DSPCOMPLEX (0, 0));
	}
}

void	fmProcessor::setfmRdsSelector (int8_t m) {
	rdsModus	= m;
}

void	fmProcessor::resetRds	(void) {
	myRdsDecoder	-> reset ();
}

void	fmProcessor::set_LocalOscillator (int32_t lo) {
	lo_frequency = lo;
}

bool	fmProcessor::ok		(void) {
	return running;
}

void	fmProcessor::setFreezer (bool b) {
	freezer = b ? 10 : 0;
	fprintf (stderr, "frezzer set to %d\n", freezer);
}

